import json

from . import RedditApp
from .utils import resolveModelFromInput, resolveUser
from ..mixins import BaseModel, DeletableMixin, EditableMixin, OwnerMixin
from ..mixins.redditReddit import RedditAppMixin


class UserVerification(BaseModel, DeletableMixin, EditableMixin, OwnerMixin, RedditAppMixin):
    _attrTypes = {
        **BaseModel._attrTypes, 'id': 'int', 'user_id': 'str', 'redditor': 'str', 'reddit_app_id': 'int', 'extra_data': 'dict', 'owner_id': 'int'
    }

    _editableAttrs = ['userId', 'redditor', 'redditAppId', 'extraData']
    _path = '/user_verifications'
    _credmgrCallable = 'userVerification'
    _nameAttr = 'userId'
    _nameMapping = {'user_id': 'userId'}
    _canFetchByName = True
    _getByNamePath = 'get_redditor'
    _apiNameMapping = {'userId': 'user_id'}

    def __init__(self, credmgr, **kwargs):
        '''Initialize an User Verification instance.

        User Verifications for associating an unique ID with a Redditor

        :param credmgr: An instance of :class:`~.CredentialManager`.
        :param id: ID of this User Verification.
        :param userId: An unique ID to associate with a Redditor.
        :param redditor: Redditor username to associate with an unique ID.
        :param redditAppId: ID of the RedditApp the Refresh Token is for.
        :param extraData: Extra data to be associated with this User Verification
        :param ownerId: ID of the `~.User` that owns this User Verification
        '''
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @resolveUser()
    def _create(_credmgr, userId, redditApp, redditor=None, extraData=None, owner=None):
        '''Create a new User Verification

        User Verifications for associating an unique ID with a Redditor

        :param str userId: An unique ID to associate with a Redditor (required)
        :param Union[RedditApp,int,str] redditApp: Reddit app the User Verification is for (required)
        :param str redditor: Redditor the User Verification is for. This is not usually set manually.
        :param dict extraData: Extra JSON data to include with verification
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to create for other users.
        :return: UserVerification
        '''

        data = {'user_id': userId, 'reddit_app_id': resolveModelFromInput(_credmgr, RedditApp, redditApp)}
        if redditor:
            data['redditor'] = redditor
        if extraData:
            data['extra_data'] = json.dumps(extraData)
        if owner:
            data['owner_id'] = owner
        return _credmgr.post('/user_verifications', data=data)


from credmgr.models.utils import camelToSnake


class EditableMixin:
    _editableAttrs = []

    def edit(self, **kwargs):
        '''Edit the object

        :param kwargs:
        :return: The edited object
        '''
        payload = []

        for attr in self._editableAttrs:
            if attr in kwargs:
                if attr in self._apiNameMapping:
                    path = f'/{self._apiNameMapping[attr]}'
                else:
                    path = f'/{camelToSnake(attr)}'
                payload.append({'op': 'replace', 'path': path, 'value': kwargs[attr]})
        self.__dict__.update(self._credmgr.patch(f'{self._path}/{self.id}', data=payload).__dict__)
        return self


from credmgr.models.utils import CachedProperty


class RedditAppMixin:
    _editableAttrs = []

    @CachedProperty
    def redditApp(self):
        if not self._fetched:
            self._fetch()
        redditApp = self._credmgr.redditApp(id=self.redditAppId)
        return redditApp


import configparser
import os
from requests import Session

from . import models
from .auth import ApiTokenAuth
from .config import Config
from .exceptions import InitializationError
from .models.utils import CachedProperty
from .requestor import Requestor, urljoin
from .serializer import Serializer


User = models.User
Bot = models.Bot
RedditApp = models.RedditApp
RefreshToken = models.RefreshToken
UserVerification = models.UserVerification
SentryToken = models.SentryToken
DatabaseCredential = models.DatabaseCredential

class CredentialManager(object):
    '''The CredentialManager class provides convenient access to CredentialManager's API.

    Instances of this class are the gateway to interacting with CredentialManager's API
    through CredentialManager. The canonical way to obtain an instance of this class is via:


    .. code-block:: python

        import CredentialManager
        credentialManager = CredentialManager.client(apiToken='LIqbGjAeep3Ws5DH3LOEQPmw8UZ6ek')


    '''
    _default = None
    _endpoint = '/api/v1'

    def __init__(self, configName=None, sessionClass=None, sessionKwargs=None, **kwargs):
        '''Initialize a CredentialManager instance.

        :param str configName: The name of a section in your ``.credmgr.ini`` file that credmgr will load its configuration
            is loaded from. If ``configName`` is ``None``, then it will look for it in the environment variable ``credmgr_configName``.
            If it is not found there, the ``default`` section is used.
        :param Session sessionClass: A Session class that will be used to create a requestor. If not set, use ``requests.Session`` (default: None).
        :param dict sessionKwargs: Dictionary with additional keyword arguments used to initialize the session (default: None).

        Additional keyword arguments will be used to initialize the
        :class:`.Config` object. This can be used to specify configuration
        settings during instantiation of the :class:`.CredentialManager` instance. For
        more details please see :ref:`configuration`.

        Required settings are:

        - apiToken

        OR

        - username

        - password

        .. warning::
             Using an API Token instead of a username/password is strongly recommended!

        The ``sessionClass`` and ``sessionKwargs`` allow for
        customization of the session :class:`.CredentialManager` will use. This allows,
        e.g., easily adding behavior to the requestor or wrapping its
        |Session|_ in a caching layer. Example usage:

        .. |Session| replace:: ``Session``
        .. _Session: https://2.python-requests.org/en/master/api/#requests.Session

        .. code-block:: python

           import json, betamax, requests

           class JSONDebugRequestor(Requestor):
               def request(self, *args, **kwargs):
                   response = super().request(*args, **kwargs)
                   print(json.dumps(response.json(), indent=4))
                   return response

           mySession = betamax.Betamax(requests.Session())
           credentialManager = CredentialManager(..., sessionClass=JSONDebugRequestor, sessionKwargs={'session': mySession})
        '''
        if sessionKwargs is None:
            sessionKwargs = {}
        configSection = None
        try:
            configSection = configName or os.getenv('credmgr_configName') or 'DEFAULT'
            self.config = Config(configSection, **kwargs)
        except configparser.NoSectionError as exc:
            if configSection is not None:
                exc.message += "\nYou provided the name of a .credmgr.ini section that doesn't exist."
            raise

        self._server = urljoin(getattr(self.config, 'server'), getattr(self.config, 'endpoint'))
        apiToken = getattr(self.config, 'apiToken')
        username = getattr(self.config, 'username')
        password = getattr(self.config, 'password')

        initErrorMessage = "Required settings are missing. Either 'apiToken' OR 'username' and 'password' must be specified, These settings can be provided in a .credmgr.ini file, as a keyword argument during the initialization of the `CredentialManager` class, or as an environment variable."
        if all([apiToken, username, password]):
            raise InitializationError(initErrorMessage)
        if apiToken:
            self._auth = ApiTokenAuth(apiToken)
        elif username and password:
            self._auth = (username, password)
        else:
            raise InitializationError('API Token or an username/password pair must be set.')

        self._requestor = Requestor(self._server, self._auth, sessionClass, **sessionKwargs)
        self.serializer = Serializer(self)
        self._currentUser = None
        self._userDefaults = None
        self.getUserDefault = lambda key, default: self.userDefaults.get(key, default)
        self.user = models.UserHelper(self)
        '''An instance of :class:`.UserHelper`.

        Provides an interface for interacting with :class:`.User`.
        For example, to get a :class:`~.User` with :attr:`id` of ``1`` you can do:
        
        .. code-block:: python
        
            user = credmgr.user(1)
            print(user.id)
        
        To create a :class:`.User` do:
        
        .. code-block:: python
        
            user = credmgr.user.create(**userKwargs)
        
        See :meth:`~.UserHelper.create` for the required params.
        '''
        self.bot = models.BotHelper(self)
        '''An instance of :class:`.BotHelper`.

        Provides an interface for interacting with :class:`.Bot`.
        For example, to get a :class:`~.Bot` with :attr:`id` of ``1`` you can do:
        
        .. code-block:: python
        
            bot = credmgr.bot(1)
            print(bot.id)
        
        To create a :class:`.Bot` do:
        
        .. code-block:: python
        
            bot = credmgr.bot.create(**botKwargs)
        
        See :meth:`~.BotHelper.create` for the required params.
        '''

        self.redditApp = models.RedditAppHelper(self)
        '''An instance of :class:`.RedditAppHelper`.

        Provides an interface for interacting with :class:`.RedditApp`.
        For example, to get a :class:`~.RedditApp` with :attr:`id` of ``1`` you can do:
        
        .. code-block:: python
        
            redditApp = credmgr.redditApp(1)
            print(redditApp.id)
        
        To create a :class:`.RedditApp` do:
        
        .. code-block:: python
        
            redditApp = credmgr.redditApp.create(**redditAppKwargs)
        
        See :meth:`~.RedditAppHelper.create` for the required params.
        '''

        self.refreshToken = models.RefreshTokenHelper(self)
        '''An instance of :class:`.RefreshTokenHelper`.

        Provides an interface for interacting with :class:`.RefreshToken`.
        For example to get a :class:`.RefreshToken` with :attr:`id` of ``1`` you can do:
        
        .. code-block:: python
        
            refreshToken = credmgr.refreshToken(1)
            print(refreshToken.id)
        
        .. note::
            Refresh tokens cannot be manually created.
        '''

        self.userVerification = models.UserVerificationHelper(self)
        '''An instance of :class:`.UserVerificationHelper`.

        Provides an interface for interacting with :class:`.UserVerification`.
        For example to get a :class:`.UserVerification` with :attr:`id` of ``1`` you can do:
        
        .. code-block:: python
        
            userVerification = credmgr.userVerification(1)
            print(userVerification.id)
        
        To create a :class:`.UserVerification` do:
        
        .. code-block:: python
        
            userVerification = credmgr.userVerification.create(**userVerificationKwargs)
        
        See :meth:`~.UserVerificationHelper.create` for the required params.
        '''

        self.sentryToken = models.SentryTokenHelper(self)
        '''An instance of :class:`.SentryTokenHelper`.

        Provides an interface for interacting with :class:`.SentryToken`.
        For example to get a :class:`.SentryToken` with :attr:`id` of ``1`` you can do:
        
        .. code-block:: python
        
            sentryToken = credmgr.sentryToken(1)
            print(sentryToken.id)
        
        To create a :class:`.SentryToken` do:
        
        .. code-block:: python
        
            sentryToken = credmgr.sentryToken.create(**sentryTokenKwargs)
        
        See :meth:`~.SentryTokenHelper.create` for the required params.
        '''

        self.databaseCredential = models.DatabaseCredentialHelper(self)
        '''An instance of :class:`.DatabaseCredentialHelper`.

        Provides an interface for interacting with :class:`.DatabaseCredential`.
        For example to get a :class:`.DatabaseCredential` with :attr:`id` of ``1`` you can do:
        
        .. code-block:: python

            databaseCredential = credmgr.databaseCredential(1)
            print(databaseCredential.id)
        
        To create a :class:`.DatabaseCredential` do:
        
        .. code-block:: python

            databaseCredential = credmgr.databaseCredential.create(**databaseCredentialKwargs)
        
        See :meth:`~.DatabaseCredentialHelper.create` for the required params.
        '''

    def users(self, batchSize=10, limit=None):
        '''List Users

        :param int batchSize: Number of Users to return in each batch (default: ``20``)
        :param int limit: Maximum number of Users to return
        :return Paginator: A :class:`~.Paginator` to iterate through the Users
        '''
        return User(self).listItems(batchSize=batchSize, limit=limit)

    def bots(self, batchSize=20, limit=None, owner=None):
        '''List Bots

        :param int batchSize: Number of Bots to return in each batch (default: ``20``)
        :param int limit: Maximum number of Bots to return
        :param Union[int,str,User] owner: Return Bots that are owner by this user
        :return Paginator: A :class:`~.Paginator` to iterate through the Bots
        '''
        return Bot(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def redditApps(self, batchSize=20, limit=None, owner=None):
        '''List RedditApps

        :param int batchSize: Number of RedditApps to return in each batch (default: ``20``)
        :param int limit: Maximum number of RedditApps to return
        :param Union[int,str,User] owner: Return RedditApps that are owner by this user
        :return Paginator: A :class:`~.Paginator` to iterate through the RedditApps
        '''
        return RedditApp(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def refreshTokens(self, batchSize=20, limit=None, owner=None):
        '''List RefreshTokens

        :param int batchSize: Number of RefreshTokens to return in each batch (default: ``20``)
        :param int limit: Maximum number of RefreshTokens to return
        :param Union[int,str,User] owner: Return RefreshTokens that are owned by this user
        :return Paginator: A :class:`~.Paginator` to iterate through the RefreshTokens
        '''
        return RefreshToken(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def userVerifications(self, batchSize=20, limit=None, owner=None):
        '''List UserVerifications

        :param int batchSize: Number of UserVerifications to return in each batch (default: ``20``)
        :param int limit: Maximum number of UserVerifications to return
        :param Union[int,str,User] owner: Return UserVerifications that are owned by this user
        :return Paginator: A :class:`~.Paginator` to iterate through the UserVerifications
        '''
        return UserVerification(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def sentryTokens(self, batchSize=20, limit=None, owner=None):
        '''List SentryTokens

        :param int batchSize: Number of SentryTokens to return in each batch (default: ``20``)
        :param int limit: Maximum number of SentryTokens to return
        :param Union[int,str,User] owner: Return SentryTokens that are owned by this user
        :return Paginator: A :class:`~.Paginator` to iterate through the SentryTokens
        '''
        return SentryToken(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def databaseCredentials(self, batchSize=20, limit=None, owner=None):
        '''List DatabaseCredentials

        :param int batchSize: Number of DatabaseCredentials to return in each batch (default: ``20``)
        :param int limit: Maximum number of DatabaseCredentials to return
        :param Union[int,str,User] owner: Return DatabaseCredentials that are owned by this user
        :return Paginator: A :class:`~.Paginator` to iterate through the DatabaseCredentials
        '''
        return DatabaseCredential(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    @CachedProperty
    def currentUser(self) -> User:
        '''Returns the currently authenticated :class:`~.User`'''
        if not self._currentUser:
            self._currentUser = self.get('/users/me')
        return self._currentUser

    @CachedProperty
    def userDefaults(self):
        '''Returns the currently authenticated :class:`~.User`'s default settings'''
        if not self._userDefaults:
            self._userDefaults = self.currentUser.defaultSettings
        return self._userDefaults

    def get(self, path, params=None):
        return self.serializer.deserialize(self._requestor.request(path, 'GET', params=params))

    def post(self, path, data):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        return self.serializer.deserialize(self._requestor.request(path, 'POST', data=data, headers=headers))

    def patch(self, path, data):
        return self.serializer.deserialize(self._requestor.request(path, 'PATCH', json=data))

    def delete(self, path):
        return self._requestor.request(path, 'DELETE')


from credmgr.mixins import BaseModel, DeletableMixin, OwnerMixin
from credmgr.mixins.redditReddit import RedditAppMixin


class RefreshToken(BaseModel, DeletableMixin, OwnerMixin, RedditAppMixin):
    _attrTypes = {
        **BaseModel._attrTypes,
        'reddit_app_id': 'int',
        'redditor': 'str',
        'refresh_token': 'str',
        'owner_id': 'int',
        'scopes': 'str',
        'issued_at': 'datetime',
        'revoked': 'bool',
        'revoked_at': 'datetime'
    }

    _path = '/refresh_tokens'
    _credmgrCallable = 'refreshToken'
    _nameMapping = {'redditor': 'redditor'}
    _apiNameMapping = {'redditor': 'redditor', 'redditAppId': 'reddit_app_id'}
    _getByNamePath = 'by_redditor'
    _canFetchByName = True
    _nameAttr = 'redditor'

    def __init__(self, credmgr, **kwargs):
        '''Initialize a Refresh Token instance.

        Refresh Tokens are for authenticating with Reddit as a Reddit that has authorized a `~.RedditApp` to access their Reddit account.

        :param credmgr: An instance of :class:`~.CredentialManager`.
        :param int id: ID of this Refresh Token.
        :param str redditor: Redditor this Refresh Token is for.
        :param str redditAppId: ID of the `~.RedditApp` this Refresh Token is for.
        :param str refreshToken: The Refresh Token to pass to a Reddit instance.
        :param list[str] scopes: The OAuth2 scopes this Refresh Token grants access to.
        :param int ownerId: ID of the `~.User` that owns this Refresh Token.
        :param datetime.datetime issuedAt: Date and time this Refresh Token was issued.
        :param bool revoked: Indicates if this Refresh Token was revoked or superseded by another Refresh Token.
        :param datetime.datetime revokedAt: Date and time this Refresh Token was revoked or superseded.
        '''
        super().__init__(credmgr, **kwargs)


'''Provides the code to load credmgr's configuration file `~/.credmgr.ini`.'''
import configparser
import os
import sys
from pathlib import Path
from threading import Lock


class Config:
    _config = None
    _lock = Lock()

    @classmethod
    def _loadConfig(cls):
        '''Attempt to load settings from various .credmgr.ini files.'''
        config = configparser.ConfigParser()
        appendFileName = lambda dirName: os.path.join(dirName, '.credmgr.ini')
        rootDir = os.path.dirname(sys.modules[__name__].__file__)
        locations = [appendFileName(rootDir), appendFileName(Path.home()), '.credmgr.ini']
        config.read(locations)
        cls._config = config

    def __init__(self, configName, **kwargs):
        '''Initialize a Config instance.'''
        with Config._lock:
            if Config._config is None:
                self._loadConfig()

        self._settings = kwargs
        self.customSettings = dict(Config._config.items(configName), **{k.lower(): v for k,v in kwargs.items()})
        self.server = self.apiToken = self.username = self.password = None

        self._initializeConfig()

    def _fetch(self, key, default=None):
        configFileValue =  self.customSettings.pop(key.lower(), default)
        envOverride = os.getenv(f'credmgr_{key}')
        return envOverride or configFileValue

    def _initializeConfig(self):
        self.server = self._fetch('server')
        self.endpoint = self._fetch('endpoint')
        self.apiToken = self._fetch('apitoken')
        self.username = self._fetch('username')
        self.password = self._fetch('password')
        self.dateFormat = self._fetch('dateformat')


from datetime import datetime

from credmgr.models.utils import resolveUser
from ..exceptions import InitializationError

class BaseModel(object):
    _attrTypes = {'id': 'int'}
    _path = None
    _nameAttr = 'name'
    _nameMapping = None
    _canFetchByName = False
    _credmgrCallable = None
    _getByNamePath = 'by_name'
    _apiNameMapping = None

    def __init__(self, credmgr, **kwargs):
        self._credmgr = credmgr
        if kwargs:
            for attribute, value in kwargs.items():
                setattr(self, attribute, value)
        self._fetched = False

    def __getattr__(self, attribute): # pragma: no cover
        '''Return the value of `attribute`.'''
        if not attribute.startswith('_') and not self._fetched:
            self._fetch()
            return getattr(self, attribute)
        raise AttributeError(f'{self.__class__.__name__!r} object has no attribute {attribute!r}')

    def _get(self, id):
        self.__dict__ = self._credmgr.get(f'{self._path}/{id}').__dict__

    def _getByName(self):
        data = {}
        for modelAttr, dataAttr in self._apiNameMapping.items():
            name = getattr(self, modelAttr, None)
            if not name: # pragma: no cover
                raise InitializationError(f'Missing required keyword arg, {modelAttr!r}, to fetch object by name')
            data[dataAttr] = name
        self.__dict__ = self._credmgr.post(f'{self._path}/{self._getByNamePath}', data=data).__dict__

    def _fetch(self, byName=False):
        if byName and self._canFetchByName:
            self._getByName()
        else:
            self._get(self.id)
        self._fetched = True

    @resolveUser()
    def listItems(self, batchSize=20, limit=None, owner=None):
        from credmgr.models.helpers import Paginator
        return Paginator(self._credmgr, self.__class__, batchSize=batchSize, limit=limit, itemsOwner=owner)

    def toDict(self):
        result = {}

        for exportAttr in self._attrTypes.keys():
            if exportAttr in self._nameMapping:
                attr = self._nameMapping.get(exportAttr)
            else:
                attr = ''.join(exportAttr.split('_')[:1] + [i.capitalize() for i in exportAttr.split('_')[1:]])
            value = getattr(self, attr, None)
            if value:
                if isinstance(value, list):
                    result[exportAttr] = list(map(lambda x: x.toDict() if hasattr(x, 'toDict') else x, value))
                elif hasattr(value, 'toDict'):
                    result[exportAttr] = value.toDict()
                elif isinstance(value, dict):
                    result[exportAttr] = dict(map(lambda item: (item[0], item[1].toDict()) if hasattr(item[1], 'toDict') else item, value.items()))
                elif isinstance(value, datetime):
                    result[exportAttr] = value.astimezone().strftime(self._credmgr.config.dateFormat)
                else:
                    result[exportAttr] = value
        if issubclass(type(self), dict):  # pragma: no cover
            for key, value in self.items():
                result[key] = value
        return result

    def __repr__(self):  # pragma: no cover
        return f'<{self.__class__.__name__} id={self.id}, {self._nameAttr}={getattr(self, self._nameAttr)!r}>'

    def __eq__(self, other):
        '''Returns true if both objects are equal'''
        if not isinstance(other, type(self)):
            return False

        return self.id == other.id and getattr(self, self._nameAttr) == getattr(other, other._nameAttr)


from .utils import resolveUser
from ..mixins import BaseApp


class DatabaseCredential(BaseApp):
    _attrTypes = {
        **BaseApp._attrTypes,
        'database_username': 'str',
        'database_host': 'str',
        'database': 'str',
        'database_flavor': 'str',
        'database_port': 'int',
        'database_password': 'str',
        'use_ssh': 'bool',
        'ssh_host': 'str',
        'ssh_port': 'int',
        'ssh_username': 'str',
        'ssh_password': 'str',
        'use_ssh_key': 'bool',
        'private_key': 'str',
        'private_key_passphrase': 'str'
    }

    _editableAttrs = BaseApp._editableAttrs + ['databaseUsername', 'databaseHost', 'database', 'databaseFlavor', 'databasePort', 'databasePassword', 'useSsh', 'sshHost', 'sshPort',
                                               'sshUsername', 'sshPassword', 'useSshKey', 'privateKey', 'privateKeyPassphrase']
    _path = '/database_credentials'
    _credmgrCallable = 'databaseCredential'

    def __init__(self, credmgr, **kwargs):
        '''Initialize a Database Credential instance.

        :param credmgr: An instance of :class:`~.CredentialManager`.
        :param id:
        :param str name: Name of the Database Credential.
        :param str databaseFlavor: Type of database.
        :param str database: Working database to use.
        :param str databaseHost: Database server address.
        :param int databasePort: Port the database server listens on.
        :param str databaseUsername: Username used to connect and authenciate the database.
        :param str databasePassword: Password used to connect and authenciate the database.
        :param bool useSsh: Determines if the database will be connected to through a tunnel.
        :param str sshHost: The address of the server that the SSH tunnel will connect to.
        :param str sshPort: The port the SSH tunnel will use.
        :param str sshUsername: Username for the SSH tunnel.
        :param str sshPassword: Password for the SSH tunnel.
        :param bool useSshKey: Determines if the SSH tunnel will use private key authenciation.
        :param str privateKey: SSH private key. Note: No validation will be performed.
        :param str privateKeyPassphrase: Passphrase for the SSH key.
        :param ownerId: ID of the `~.User` that owns this Database Credential.
        '''
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @resolveUser()
    def _create(_credmgr, name, databaseFlavor='postgres', database='postgres', databaseHost='localhost', databasePort=5432, databaseUsername='postgres', databasePassword=None,
            useSsh=False, sshHost=None, sshPort=None, sshUsername=None, sshPassword=None, useSshKey=False, privateKey=None, privateKeyPassphrase=None, enabled=True, owner=None):
        '''Create a new Database Credential

        Database Credentials are used for..ya know..databases

        :param str name: Name of the Database Credential (required)
        :param str databaseFlavor: Type of database, (default: ``postgres``)
        :param str database: Working database to use, (default: ``postgres``)
        :param str databaseHost: Database server address, (default: ``localhost``)
        :param int databasePort: Port the database server listens on, (default: ``5432``)
        :param str databaseUsername: Username to use to connect to the database
        :param str databasePassword: Password to use to connect to the database
        :param bool useSsh: Determines if the database will be connected to through a tunnel
        :param str sshHost: The address of the server that the SSH tunnel will connect to
        :param str sshPort: The port the SSH tunnel will use
        :param str sshUsername: Username for the SSH tunnel
        :param str sshPassword: Password for the SSH tunnel
        :param bool useSshKey: Allows the credentials to be used
        :param str privateKey: SSH private key. Note: No validation will be performed.
        :param str privateKeyPassphrase: Passphrase for the SSH key
        :param bool enabled: Allows the credentials to be used
        :param Union[User,int,str] owner: Owner of the Database Credential. Requires Admin to create for other users.
        :returns: DatabaseCredential
        '''

        data = {}
        if databaseFlavor:
            data['database_flavor'] = databaseFlavor
        if database:
            data['database'] = database
        if databaseHost:
            data['database_host'] = databaseHost
        if databasePort:
            data['database_port'] = databasePort
        if databaseUsername:
            data['database_username'] = databaseUsername
        if databasePassword:
            data['database_password'] = databasePassword
        if useSsh:
            data['use_ssh'] = useSsh
        if sshHost:
            data['ssh_host'] = sshHost
        if sshPort:
            data['ssh_port'] = sshPort
        if sshUsername:
            data['ssh_username'] = sshUsername
        if sshPassword:
            data['ssh_password'] = sshPassword
        if useSshKey:
            data['use_ssh_key'] = useSshKey
        if privateKey:
            data['private_key'] = privateKey
        if privateKeyPassphrase:
            data['private_key_passphrase'] = privateKeyPassphrase
        if enabled:
            data['enabled'] = enabled
        if owner:
            data['owner_id'] = owner
        return _credmgr.post('/database_credentials', data={'app_name': name, **data})


import json

from ..exceptions import InitializationError
from ..mixins import BaseModel, DeletableMixin, EditableMixin


class User(BaseModel, DeletableMixin, EditableMixin):
    _attrTypes = {
        **BaseModel._attrTypes,
        'id': 'int',
        'username': 'str',
        'is_active': 'bool',
        'is_regular_user': 'bool',
        'is_admin': 'bool',
        'default_settings': 'dict(str, str)',
        'reddit_username': 'str',
        'created': 'datetime',
        'updated': 'datetime',
        'reddit_apps': 'list[RedditApp]',
        'sentry_tokens': 'list[SentryToken]',
        'database_credentials': 'list[DatabaseCredential]'
    }
    _editableAttrs = ['username', 'isActive', 'isRegularUser', 'isAdmin', 'defaultSettings', 'redditUsername']
    _path = '/users'
    _credmgrCallable = 'user'
    _nameAttr = 'username'
    _nameMapping = {'username': 'username'}
    _canFetchByName = True
    _apiNameMapping = {_nameAttr: 'username'}

    def __init__(self, credmgr, **kwargs):
        '''Initialize an User instance

        Users are for logging into and accessing that user's credentials

        :param credmgr: An instance of :class:`~.CredentialManager`.
        :param int id: ID of the User.
        :param username: Username of the User.
        :param bool isActive: Indicates if the User can login and access CredentialManager.
        :param bool isRegularUser: Incicates if this is a regular user.
        :param bool isAdmin: Incicates if this User can create other users and their credentials.
        :param defaultSettings:
        :param str redditUsername: This User's Reddit username. Used for :class:`~.RedditApp`'s userAgent.
        :param datetime.datetime created: Date and time this User was created.
        :param datetime.datetime updated: Date and time this User was last updated.
        :param list[RedditApp] redditApps: A list of Reddit Apps this User owns.
        :param list[SentryToken] sentryTokens: A list of Sentry Tokens this User owns.
        :param list[DatabaseCredential] databaseCredentials: A list of Database Credentials this User owns.
        '''
        super().__init__(credmgr, **kwargs)
        self._apps = {}
        if 'redditApps' in kwargs:
            self._apps['redditApps'] = kwargs['redditApps']
        if 'sentryTokens' in kwargs:
            self._apps['sentryTokens'] = kwargs['sentryTokens']
        if 'databaseCredentials' in kwargs:
            self._apps['databaseCredentials'] = kwargs['databaseCredentials']

    @staticmethod
    def _create(_credmgr, username, password, defaultSettings=None, redditUsername=None, isAdmin=False, isActive=True, isRegularUser=True, isInternal=False):
        '''Create a new User

        **PERMISSIONS: Admin role is required.**

        :param str username: Username for new user (Example: ```spaz```) (required)
        :param str password: Password for new user (Example: ```supersecurepassword```) (required)
        :param dict defaultSettings: Default values to use for new apps (Example: ```{"databaseFlavor": "postgres", "databaseHost": "localhost"}```)
        :param str redditUsername: User's Reddit username (Example: ```LilSpazJoekp```)
        :param bool isAdmin: Is the user an admin? Allows the user to see all objects and create users (Default: ``false``)
        :param bool isActive: Is the user active? Allows the user to sign in (Default: ``true``)
        :param bool isRegularUser: (Internal use only)
        :param bool isInternal: (Internal use only)
        :return: User

        '''
        additionalParams = {}
        if defaultSettings:
            additionalParams['default_settings'] = json.dumps(defaultSettings)
        if isAdmin: # pragma: no cover
            additionalParams['is_admin'] = isAdmin
        if isActive:
            additionalParams['is_active'] = isActive
        if isRegularUser:
            additionalParams['is_regular_user'] = isRegularUser
        if isInternal: # pragma: no cover
            additionalParams['is_internal'] = isInternal
        if redditUsername:
            additionalParams['reddit_username'] = redditUsername
        return _credmgr.post('/users', data={'username': username, 'password': password, **additionalParams})

    def apps(self, only=None):
        '''Returns apps that are owned by this user

        :param str only: Pass one of ``redditApps``, ``sentryTokens``, ``databaseCredentials`` to only get that type of apps
        :return Union[dict[str,list[Union[RedditApp,SentryToken,DatabaseCredential]]],list[Union[RedditApp,SentryToken,DatabaseCredential]]]:
        '''
        appTypes = ['redditApps', 'sentryTokens', 'databaseCredentials']
        if not self._apps:
            response = self._credmgr.get(f'/users/{self.id}/apps')
            self._apps = response._apps
            for app in appTypes:
                setattr(self, app, self._apps[app])
        if only:
            if only in appTypes:
                return self._apps[only]
            else:
                raise InitializationError(f"App type: {only} is not valid. Only 'redditApps', 'sentryTokens', and 'databaseCredentials' are valid.")
        return self._apps


from ..mixins import BaseModel, DeletableMixin, EditableMixin, OwnerMixin


class BaseApp(BaseModel, DeletableMixin, EditableMixin, OwnerMixin):
    _attrTypes = {**BaseModel._attrTypes, 'app_name': 'str', 'owner_id': 'int'}
    _nameAttr = 'name'
    _nameMapping = {'app_name': 'name'}
    _editableAttrs = ['name',]
    _apiNameMapping = {'name': 'app_name'}


from .utils import resolveUser
from ..mixins import BaseApp


class SentryToken(BaseApp):
    _attrTypes = {**BaseApp._attrTypes, 'dsn': 'str'}
    _editableAttrs = BaseApp._editableAttrs + ['dsn']
    _path = '/sentry_tokens'
    _credmgrCallable = 'sentryToken'

    def __init__(self, credmgr, **kwargs):
        '''Initialize a Sentry Token instance.

        :param credmgr: An instance of :class:`~.CredentialManager`.
        :param id: ID of the Sentry Token
        :param str name: Name of the Sentry Token.
        :param str dsn: Sentry DSN url for reporting errors.
        :param ownerId: ID of the `~.User` that owns this Sentry Token.
        '''
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @resolveUser()
    def _create(_credmgr, name=None, dsn=None, owner=None):
        '''Create a new Sentry Token

        Sentry Tokens are used for logging and error reporting in applications

        :param str name: Name of the Sentry Token (required)
        :param str dsn: DSN of the Sentry Token (required)
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to create for other users.
        :return: SentryToken
        '''
        data = {'app_name': name, 'dsn': dsn}
        if owner:
            data['owner_id'] = owner
        return _credmgr.post('/sentry_tokens', data=data)




from .utils import camelToSnake, resolveUser
from ..mixins import BaseApp


class Bot(BaseApp):
    _attrTypes = {
        **BaseApp._attrTypes, 'reddit_app': 'RedditApp', 'sentry_token': 'SentryToken', 'database_credential': 'DatabaseCredential'
    }
    _editableAttrs = BaseApp._editableAttrs + ['redditAppId', 'sentryTokenId', 'databaseCredentialId']
    _path = '/bots'
    _credmgrCallable = 'bot'
    _canFetchByName = True

    def __init__(self, credmgr, **kwargs):
        '''Initialize a Bot instance

        Bots are used for grouping apps into a single request

        :param credmgr: An instance of :class:`~.CredentialManager`.
        :param id: ID of this Bot.
        :param name: Name of this Bot.
        :param ownerId: ID of the `~.User` that owns this Bot.
        :param redditApp: `~.RedditApp` that will be used with this Bot.
        :param sentryToken: `~.SentryToken` that will be used with this Bot.
        :param databaseCredential: `~.DatabaseCredential` that will be used with this Bot.
        '''
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @resolveUser()
    def _create(_credmgr, name, redditApp, sentryToken, databaseCredential, owner=None):
        '''Create a new Bot

        Bots are used for grouping apps into a single request

        :param str name: Name of the Bot (required)
        :param Union[RedditApp,int] redditApp: Reddit App the bot will use
        :param Union[SentryToken,int] sentryToken: Sentry Token the bot will use
        :param Union[DatabaseCredential,int] databaseCredential: Database Credentials the bot will use
        :param Union[User,int,str] owner: Owner of the bot. Requires Admin to create for other users.
        :return: Bot
        '''

        from . import DatabaseCredential, RedditApp, SentryToken
        additionalParams = {}
        if isinstance(redditApp, RedditApp):
            redditApp = redditApp.id
        if redditApp:
            additionalParams['reddit_app_id'] = redditApp
        if isinstance(sentryToken, SentryToken):
            sentryToken = sentryToken.id
        if sentryToken:
            additionalParams['sentry_token_id'] = sentryToken
        if isinstance(databaseCredential, DatabaseCredential):
            databaseCredential = databaseCredential.id
        if databaseCredential:
            additionalParams['database_credential_id'] = databaseCredential
        if owner:
            additionalParams['owner_id'] = owner
        return _credmgr.post('/bots', data={'app_name': name, **additionalParams})

    def edit(self, **kwargs):
        '''

        :param name: Changes the name of the :class:`~.Bot`
        :param Union[RedditApp,int] redditApp: Changes the :class:`~.RedditApp` the Bot will use
        :param Union[SentryToken,int] sentryToken: Changes the :class:`~.SentryToken` the Bot will use
        :param Union[DatabaseCredential,int] databaseCredential: Changes the :class:`~.DatabaseCredential` the Bot will use

        .. note ::
            Parameters, ``redditApp``, ``sentryToken``, and ``databaseCredential`` can accept the initialized object or its :attr:`id`.

        :return: The modified :class:`~.Bot`
        '''
        from credmgr.models import DatabaseCredential, RedditApp, SentryToken
        iterKwargs = dict(kwargs.items())
        for key, value in iterKwargs:
            if key in ['redditApp', 'sentryToken', 'databaseCredential']:
                if isinstance(kwargs[key], (RedditApp, SentryToken, DatabaseCredential)):
                    newKey = f'{key}Id'
                    kwargs[newKey] = kwargs.pop(key).id
                else:
                    newKey = f'{key}Id'
                    kwargs[newKey] = kwargs.pop(key)
        super(Bot, self).edit(**kwargs)


from credmgr.models.utils import CachedProperty


class OwnerMixin:
    _editableAttrs = []

    @CachedProperty
    def owner(self):
        if not self._fetched:
            self._fetch()
        user = self._credmgr.user(id=self.ownerId)
        return user